import os
import math
import copy
import numpy as np
from fpdf import FPDF
from datetime import datetime
import matplotlib.pyplot as plt
from sklearn.model_selection import StratifiedKFold
from sklearn.metrics import plot_roc_curve
from sklearn.metrics import auc
from urllib.error import URLError


class BinaryDocumenting(FPDF):

    def __init__(self, pipeline):
        super().__init__()

        # Settings
        self.WIDTH = 210
        self.HEIGHT = 297
        self.lh = 6  # Line Height
        self.pm = 15  # Paper margin
        self.tm = 8  # Text Margins
        self.set_margins(self.pm, self.pm, self.pm)

        # Args
        self.p = pipeline
        self.name = pipeline.name

        # Initiates
        self.model = None
        self.mName = None
        self.feature_set = None
        self.metrics = None
        self.outputMetrics = None
        self.confusion_matrix = None
        self.cv = None
        self.x = None
        self.y = None
        self.featureImportances = {}

    def header(self):
        try:        # No problem if not possible due lack of internet
            self.image('https://raw.githubusercontent.com/nielsuit227/AutoML/main/Amplo/Static/logo.png',
                       x=self.pm, y=self.pm, w=55, h=15)
        except URLError:
            pass
        self.set_font('Helvetica', '', 14)
        self.set_text_color(100, 100, 100)
        self.cell(self.WIDTH - 80)
        self.cell(0, 15, 'Amplo AutoML Documentation', 0, 0, align='R')
        self.set_font('Helvetica', '', 10)
        self.ln(6)
        self.cell(0, 15, '{}'.format(datetime.now().strftime('%d %b %Y - %H:%M')), 0, 0, align='R')
        self.ln(25)

    def create(self, model, feature_set):
        """
        Creates the entire documentation.

        Parameters
        ----------
        model : Model to document / validate
        feature_set : Feature set to document / validate on
        """
        # Asserts
        assert model is not None, 'Model cannot be none.'
        assert feature_set in self.p.featureSets.keys(), 'Feature set unavailable'

        # Set variables
        self.model = model
        self.mName = type(model).__name__
        self.feature_set = feature_set

        # Analyse the model
        self.prepare_data()
        self.analyse()

        # Check if folder needs creation
        path = self.p.mainDir + 'Documentation/v{}/{}_{}.pdf'.format(self.p.version, self.mName, feature_set)
        if not os.path.exists(path[:path.rfind('/')]):
            os.makedirs(path[:path.rfind('/')])

        # Create PDF
        self.add_page()
        self.add_h1('{}'.format(self.name))
        self.add_h2('{} - v{}'.format(self.mName, self.p.version))
        self.add_text(self.model_description)
        self.model_performance()
        self.validation()
        self.parameters()
        self.features()
        self.data()
        self.score_board()
        self.output(self.p.mainDir + 'Documentation/v{}/{}_{}.pdf'.format(self.p.version, self.mName, self.feature_set))

    def prepare_data(self):
        """
        Although data of the Pipeline (self.p.x and self.p.y), is already cleaned, it is not normalized or sequenced.
        @return:
        @rtype:
        """
        x, y = self.p.x, self.p.y
        self.x, self.y = x.to_numpy(), y.to_numpy()

    def analyse(self):
        # Metrics & Confusion Matrix
        accuracy = []
        precision = []
        sensitivity = []
        specificity = []
        f1_score = []
        area_under_curves = []
        true_positive_rates = []
        cm = np.zeros((self.p.cvSplits, 2, 2))
        mean_fpr = np.linspace(0, 1, 100)

        # Plot Initiator
        fig, ax = plt.subplots(math.ceil(self.p.cvSplits / 2), 2, sharex='all', sharey='all', figsize=[12, 8])
        fig.suptitle('{}-Fold Cross Validated Predictions - {} ({})'.format(
            self.p.cvSplits, self.mName, self.feature_set))
        fig2, ax2 = plt.subplots(figsize=[12, 8])

        # Modelling
        self.cv = StratifiedKFold(n_splits=self.p.cvSplits, shuffle=self.p.shuffle)
        for i, (t, v) in enumerate(self.cv.split(self.x, self.y)):
            n = len(v)
            xt, xv, yt, yv = self.x[t], self.x[v], self.y[t].reshape((-1)), self.y[v].reshape((-1))
            model = copy.deepcopy(self.model)
            model.fit(xt, yt)
            predictions = model.predict(xv).reshape((-1))

            # Metrics
            tp = np.logical_and(np.sign(predictions) == 1, yv == 1).sum()
            tn = np.logical_and(np.sign(predictions) == 0, yv == 0).sum()
            fp = np.logical_and(np.sign(predictions) == 1, yv == 0).sum()
            fn = np.logical_and(np.sign(predictions) == 0, yv == 1).sum()
            accuracy.append((tp + tn) / n * 100)
            if tp + fp > 0:
                precision.append(tp / (tp + fp) * 100)
            if tp + fn > 0:
                sensitivity.append(tp / (tp + fn) * 100)
            if tn + fp > 0:
                specificity.append(tn / (tn + fp) * 100)
            if tp + fp > 0 and tp + fn > 0:
                f1_score.append(2 * precision[-1] * sensitivity[-1] / (precision[-1] + sensitivity[-1]) if
                                precision[-1] + sensitivity[-1] > 0 else 0)
            cm[i] = np.array([[tp, fp], [fn, tn]])

            # ROC calculations
            viz = plot_roc_curve(model, xv, yv, name='ROC fold {}'.format(i + 1), alpha=0.3, ax=ax2)
            interp_tpr = np.interp(mean_fpr, viz.fpr, viz.tpr)
            interp_tpr[0] = 0.0
            true_positive_rates.append(interp_tpr)
            area_under_curves.append(viz.roc_auc)

            # Plot
            ax[i // 2][i % 2].plot(yv, c='#2369ec', alpha=0.6)
            ax[i // 2][i % 2].plot(predictions, c='#ffa62b')
            ax[i // 2][i % 2].set_title('Fold-{}'.format(i))

        # Statistics on results
        totals = np.sum(cm, axis=(1, 2), keepdims=True)
        means = np.mean(cm / totals * 100, axis=0).astype('float')
        stds = np.std(cm / totals * 100, axis=0).astype('float')

        # Store
        self.metrics = {
            'Accuracy': [float(np.mean(accuracy)), float(np.std(accuracy))],
            'Precision': [float(np.mean(precision)), float(np.std(precision))],
            'Sensitivity': [float(np.mean(sensitivity)), float(np.std(sensitivity))],
            'Specificity': [float(np.mean(specificity)), float(np.std(specificity))],
            'F1 Score': [float(np.mean(f1_score)), float(np.std(f1_score))],
        }
        self.confusion_matrix = {
            'means': means,
            'stds': stds,
        }
        self.outputMetrics = copy.deepcopy(self.metrics)
        self.outputMetrics['True Positives'] = [means[0, 0], stds[0, 0]]        # These are float32s
        self.outputMetrics['False Positives'] = [means[0, 1], stds[0, 1]]
        self.outputMetrics['False Negatives'] = [means[1, 0], stds[1, 0]]
        self.outputMetrics['True Negatives'] = [means[1, 1], stds[1, 1]]

        # Print Results
        print('[AutoML] Accuracy:        {:.2f} \u00B1 {:.2f} %'.format(np.mean(accuracy), np.std(accuracy)))
        print('[AutoML] Precision:       {:.2f} \u00B1 {:.2f} %'.format(np.mean(precision), np.std(precision)))
        print('[AutoML] Sensitivity:     {:.2f} \u00B1 {:.2f} %'.format(
            np.mean(sensitivity), np.std(sensitivity)))
        print('[AutoML] Specificity:     {:.2f} \u00B1 {:.2f} %'.format(
            np.mean(specificity), np.std(specificity)))
        print('[AutoML] F1-score:        {:.2f} \u00B1 {:.2f} %'.format(np.mean(f1_score), np.std(f1_score)))
        print('[AutoML] Confusion Matrix:')
        print('[AutoML] Prediction / true |     Faulty     |    Healthy      ')
        print('[AutoML]       Faulty      | {} |  {}'.format(
            ('{:.1f} \u00B1 {:.1f} %'.format(means[0, 0], stds[0, 0])).ljust(14),
            ('{:.1f} \u00B1 {:.1f} %'.format(means[0, 1], stds[0, 1])).ljust(14)))
        print('[AutoML]       Healthy     | {} |  {}'.format(
            ('{:.1f} \u00B1 {:.1f} %'.format(means[1, 0], stds[1, 0])).ljust(14),
            ('{:.1f} \u00B1 {:.1f} %'.format(means[1, 1], stds[1, 1])).ljust(14)))

        # Check whether plot is possible
        if type(self.model).__name__ == 'Lasso' or 'Ridge' in type(self.model).__name__:
            return

        # Adjust plots
        ax2.plot([0, 1], [0, 1], linestyle='--', lw=2, color='#ffa62b',
                 label='Chance', alpha=.8)
        mean_tpr = np.mean(true_positive_rates, axis=0)
        mean_tpr[-1] = 1.0
        mean_auc = auc(mean_fpr, mean_tpr)
        std_auc = np.std(area_under_curves)
        ax2.plot(mean_fpr, mean_tpr, color='#2369ec',
                 label=r'Mean ROC (AUC = {:.2f} $\pm$ {:.2f})'.format(mean_auc, std_auc),
                 lw=2, alpha=.8)
        std_tpr = np.std(true_positive_rates, axis=0)
        true_pos_rates_upper = np.minimum(mean_tpr + std_tpr, 1)
        true_pos_rates_lower = np.maximum(mean_tpr - std_tpr, 0)
        ax2.fill_between(mean_fpr, true_pos_rates_lower, true_pos_rates_upper, color='#729ce9', alpha=.2,
                         label=r'$\pm$ 1 std. dev.')
        ax2.set(xlim=[-0.05, 1.05], ylim=[-0.05, 1.05],
                title="ROC Curve - {}".format(self.mName))
        ax2.legend(loc="lower right")
        if not os.path.exists(self.p.mainDir + 'EDA/Validation/v{}/'.format(self.p.version)):
            os.makedirs(self.p.mainDir + 'EDA/Validation/v{}/'.format(self.p.version))
        roc_path = self.p.mainDir + 'EDA/Validation/v{}/ROC_{}.png'.format(self.p.version, self.mName)
        fig2.savefig(roc_path, format='png', dpi=200)
        cross_val_path = self.p.mainDir + 'EDA/Validation/v{}/Cross_Val_{}.png'.format(self.p.version, self.mName)
        fig.savefig(cross_val_path, format='png', dpi=200)

        # Feature Importance (only if EDA is not run)
        if not os.path.exists(self.p.mainDir + 'EDA/Features/v{}/RF.png'.format(self.p.version)):
            if not os.path.exists(self.p.mainDir + 'EDA/Features/v{}'.format(self.p.version)):
                os.makedirs(self.p.mainDir + 'EDA/Features/v{}/'.format(self.p.version))
            from sklearn.ensemble import RandomForestClassifier
            model = RandomForestClassifier()
            model.fit(self.p.x, self.p.y)
            fig, ax = plt.subplots(figsize=[4, 6], constrained_layout=True)
            plt.subplots_adjust(left=0.5, top=1, bottom=0)
            ax.spines['right'].set_visible(False)
            ax.spines['bottom'].set_visible(False)
            ax.spines['top'].set_visible(False)
            ind = np.argsort(model.feature_importances_)
            plt.barh(list(self.p.x.keys()[ind])[-15:], width=model.feature_importances_[ind][-15:],
                     color='#2369ec')
            fig.savefig(self.p.mainDir + 'EDA/Features/v{}/RF.png'.format(self.p.version), format='png', dpi=200)

    def check_new_page(self, margin=220):
        if self.get_y() > margin:
            self.add_page()
            return True
        else:
            return False

    def add_h1(self, title):
        self.set_font('Helvetica', 'B', 18)
        self.set_text_color(80, 80, 80)
        self.cell(0, 0, title, 0, 0)
        self.ln(self.lh * 2)

    def add_h2(self, title):
        self.set_font('Helvetica', 'B', 16)
        self.set_text_color(80, 80, 80)
        self.cell(0, 0, title, 0, 0)
        self.ln(self.lh)

    def add_h3(self, title, **args):
        self.set_font('Helvetica', 'B', 14)
        self.set_text_color(80, 80, 80)
        self.cell(0, 0, title, **args)
        self.ln(self.lh)

    def add_text(self, text):
        self.set_font('Helvetica', '', 12)
        self.set_text_color(80, 80, 80)
        self.multi_cell(0, 5, text)

    def add_image(self, path, height, width):
        self.image(path, h=height, w=width, x=self.WIDTH - width)

    @property
    def model_description(self):
        if 'CatBoost' in self.mName:
            return "CatBoost, or Categorical Boosting,  is an algorithm for gradient boosting on " \
                   "decision trees, with natural implementation for categorical variables. It is similar to " \
                   "XGBoost and LightGBM but differs in implementation of the optimization algorithm. We often " \
                   "see this algorithm performing very well."
        elif 'XGB' in self.mName:
            return "XGBoost, or Extreme Gradient Boosting, is an algorithm for gradient boosting on " \
                   "decision trees. It trains many decision trees sequentially, the additional tree always " \
                   "trying to mitigate the error of the whole model. XGBoost was the first gradient boosting " \
                   "algorithm to be implemented and is currently widely adopted in the ML world. "
        elif 'LGBM' in self.mName:
            return "LightGBM, or Light Gradient Boosting Machine, is an iteration on the XGBoost" \
                   " algorithm. Similarly, it uses gradient boosting with decision trees. However, XGBoost tend " \
                   "to be slow for a larger number of samples (>10.000), but with leaf-wise growth instead of " \
                   "depth-wise growth, LightGBM increases training speed significantly. Performance is often " \
                   "close to XGBoost, sometimes for the better and sometimes for the worse."
        elif 'HistG' in self.mName:
            return "SciKits implementation of LightGBM, or Light Gradient Boosting Machine, is" \
                   " an iteration on the XGBoost " \
                   "algorithm. Similarly, it uses gradient boosting with decision trees. However, XGBoost tend " \
                   "to be slow for a larger number of samples (>10.000), but with leaf-wise growth instead of " \
                   "depth-wise growth, LightGBM increases training speed significantly. Performance is often " \
                   "close to XGBoost, sometimes for the better and sometimes for the worse. "
        elif 'GradientB' in self.mName:
            return "SciKits implementation of XGBoost, or Extreme Gradient Boosting, is an algorithm for" \
                   " gradient boosting on " \
                   "decision trees. It trains many decision trees sequentially, the additional tree always " \
                   "trying to mitigate the error of the whole model. XGBoost was the first gradient boosting " \
                   "algorithm to be implemented and is currently widely adopted in the ML world. "
        elif 'RandomForest' in self.mName:
            return "Random Forest, is an ensemble algorithm that combines many (100-1000) decision trees and " \
                   "predicts the average of all trained trees. Though gradient boosting methods often outperform " \
                   "Random Forest, some data characteristics favor the Random Forests performance. "
        elif 'Linear' in self.mName:
            return "Linear models are simple algorithms where the inputs are multiplied by optimized weights to " \
                   "predict the output. "
        elif 'Bagging' in self.mName:
            return """Bagging algorithms is an ensemble algorithm. Just like a Random Forest, it trains many 
            Decision Trees. It then makes a prediction based on a voting base, the average of the prediction  
            of the individual Decision Trees will be predicted by the Bagging algorithm. Contrary to a Random Forest,
            the Bagging algorithm does not allocate subsets of features or data to the individual Decision Trees.
            """
        elif 'Stacking' in self.mName:
            return """A Stacking algorithm is a linear algorithm that uses the prediction of various models. In our 
            AutoML pipeline, we first train all included models, optimize the hyper parameters of the well performing 
            ones and then take the three best combinations. Additionally, we add a Naive Bayes, a Linear Model and a 
            K-Nearest Neighbors algorithm on top. 
            """
        else:
            return """Model description yet to be included in this documenter."""

    def model_performance(self):
        if not self.check_new_page():
            self.ln(self.lh)
        self.add_h2('Model Performance')
        self.add_text('Model performance is analysed by various metrics. This model has been selected based on the {} '
                      'score.'.format(self.p.objective))

        # Metrics
        self.set_font('Helvetica', 'B', 12)
        self.ln(self.lh)
        self.cell(w=50, h=self.lh, txt='Metric', border='B', align='C')
        self.cell(w=50, h=self.lh, txt='Score', border='LB', align='C')
        self.set_font('Helvetica', '', 12)
        for k, v in self.metrics.items():
            self.ln(self.lh)
            self.cell(w=50, h=self.lh, txt=k, border='R', align='L')
            self.cell(w=50, h=self.lh, txt='{:.2f} \u00B1 {:.2f} %'.format(v[0], v[1]), border='L', align='C')
        self.ln(self.lh * 3)

        # Confusion Matrix
        n_classes = 2
        cell_width = int((self.WIDTH - self.pm * 4 - self.tm * n_classes) / (n_classes + 2))
        self.add_h3('Confusion Matrix')

        # First row
        self.set_font('Helvetica', 'B', 12)
        self.cell(w=cell_width * 2, h=self.lh, txt='', align='L', border='R')
        self.cell(w=cell_width * n_classes, h=self.lh, txt='True Class', align='C')
        self.ln(self.lh)

        # Second Row
        self.cell(w=cell_width * 2, h=self.lh, txt='', align='L', border='BR')
        self.cell(w=cell_width, h=self.lh, txt='Faulty', align='C', border='B')
        self.cell(w=cell_width, h=self.lh, txt='Healthy', align='C', border='B')
        self.ln(self.lh)

        # Third Row (first with values)
        self.cell(w=cell_width, h=self.lh * 2, txt='Prediction', align='L')
        self.cell(w=cell_width, h=self.lh, txt='Faulty', align='L', border='R')
        self.set_font('Helvetica', '', 12)
        self.cell(w=cell_width, h=self.lh, txt='{:.2f} \u00B1 {:.2f} %'.format(
            self.confusion_matrix['means'][0][0], self.confusion_matrix['stds'][0][0]),
                  align='C')
        self.cell(w=cell_width, h=self.lh, txt='{:.2f} \u00B1 {:.2f} %'.format(
            self.confusion_matrix['means'][0][1], self.confusion_matrix['stds'][0][1]),
                  align='C')
        self.ln(self.lh)

        # Fourth Row
        self.set_font('Helvetica', 'B', 12)
        self.cell(w=cell_width, h=self.lh, txt='', align='L')
        self.cell(w=cell_width, h=self.lh, txt='Healthy', align='L', border='R')
        self.set_font('Helvetica', '', 12)
        self.cell(w=cell_width, h=self.lh, txt='{:.2f} \u00B1 {:.2f} %'.format(
            self.confusion_matrix['means'][1][0], self.confusion_matrix['stds'][1][0]),
                  align='C')
        self.cell(w=cell_width, h=self.lh, txt='{:.2f} \u00B1 {:.2f} %'.format(
            self.confusion_matrix['means'][1][1], self.confusion_matrix['stds'][1][1]),
                  align='C')
        self.ln(self.lh)

    def validation(self):
        if not self.check_new_page():
            self.ln(self.lh)
        self.ln(self.lh * 2)
        self.add_h3('Area Under Curve & Cross Validation Plots')
        x, y = self.get_x(), self.get_y()
        path = self.p.mainDir + 'EDA/Validation/v{}/ROC_{}.png'.format(self.p.version, self.mName)
        self.image(x=x, y=y, w=90, h=60, name=path)
        path = path[:path.rfind('/')] + '/Cross_Val_{}.png'.format(self.mName)
        self.image(x=x + self.WIDTH / 2 - self.pm, y=y, w=90, h=60, name=path)
        self.ln(self.lh)
        self.add_page()
        self.add_h3('Validation Strategy')
        self.add_text("All experiments are cross-validated. This means that every time a model's "
                      "performance is evaluated, it's trained on one part of the data, and test on another. Therefore, "
                      "the model is always test against data it has not yet been trained for. This gives the best "
                      "approximation for real world (out of sample) performance. The current validation strategy used "
                      "is {}, with {} splits and {} shuffling the data."
                      .format(type(self.cv).__name__, self.p.cvSplits, 'with' if self.p.shuffle else 'without'))
        self.ln(self.lh)

    def parameters(self):
        if not self.check_new_page():
            self.ln(self.lh)
        params = self.model.get_params()
        n_params = len(params)
        keys, values = list(params.keys()), list(params.values())
        self.ln(self.lh)
        self.add_h2('Model Parameters')
        self.set_font('Helvetica', 'B', 12)

        # Double rows
        if n_params > 15:
            # First row
            w = 30
            self.cell(w=w + 30, h=self.lh, txt='Parameter', align='C', border='RB')
            self.cell(w=w, h=self.lh, txt='Value', align='C', border='B')
            self.cell(w=5, h=self.lh, txt='')
            self.cell(w=w + 30, h=self.lh, txt='Parameter', align='C', border='RB')
            self.cell(w=w, h=self.lh, txt='Value', align='C', border='B')
            self.set_font('Helvetica', '', 12)
            n_rows = - (-n_params // 2)
            for i in range(n_rows):
                self.ln(self.lh)
                self.cell(w=w + 30, h=self.lh, txt=keys[i * 2], align='L', border='R')
                value = '{:.4e}'.format(values[i * 2]) if isinstance(values[i * 2], float) else str(values[i * 2])
                self.cell(w=w, h=self.lh, txt=value, align='C')
                self.cell(w=5, h=self.lh),
                if i * 2 + 1 < n_params:
                    self.cell(w=w + 30, h=self.lh, txt=keys[i * 2 + 1], align='L', border='R')
                    value = '{:.4e}'.format(values[i * 2 + 1]) if isinstance(values[i * 2 + 1], float) else \
                        str(values[i * 2 + 1])
                    self.cell(w=w, h=self.lh, txt=value, align='C')

        # Single rows
        else:
            w = 50
            self.cell(w=w, h=self.lh, txt='Parameter', align='L', border='RB')
            self.cell(w=w, h=self.lh, txt='Value', align='L', border='B')
            for i in range(n_params):
                self.ln(self.lh)
                self.cell(w=w, h=self.lh, txt=keys[i], align='L', border='R')
                value = '{:.4e}'.format(values[i]) if isinstance(values[i], float) else str(values[i])
                self.cell(w=w, h=self.lh, txt=value, align='R')

    def features(self):
        features = {
            'Co-Linear Features': self.p.featureProcesser.coLinearFeatures,
            'Linear Features': self.p.featureProcesser.linearFeatures,
            'Arithmetic Features': self.p.featureProcesser.crossFeatures,
            'Trigonometric Features': self.p.featureProcesser.trigonometricFeatures,
            'Inverse Features': self.p.featureProcesser.inverseFeatures,
            'K-Means Features': self.p.featureProcesser.kMeansFeatures,
            'Lagged Features': self.p.featureProcesser.laggedFeatures,
            'Differentiated Features': self.p.featureProcesser.diffFeatures,
        }
        if not self.check_new_page():
            self.ln(20)
        self.add_h2('Features')

        # Feature Extraction
        self.add_h3('Feature Extraction')
        self.add_text('First, features that are co-linear (a * x = y) up to {} % were removed. This resulted in {} '
                      'removed features: {}{}'
                      .format(self.p.informationThreshold * 100, len(features['Co-Linear Features']),
                              ', '.join([i for i in features['Co-Linear Features'][:20]]),
                              ', ...' if len(features['Co-Linear Features']) > 20 else ' ', ))
        self.check_new_page(margin=220)
        self.add_text('Subsequently, the features were manipulated and analysed to extract additional information. All '
                      'promising combinations are analysed with a single shallow decision tree.')
        features.pop('Co-Linear Features')
        self.ln()
        self.set_font('Helvetica', 'B', 12)
        self.cell(w=50, h=self.lh, txt='Sort Feature', border='BR', align='C')
        self.cell(w=50, h=self.lh, txt='Quantity', border='B', align='C')
        # Not supported now as multi cell adds whitespace in between rows, maybe appendix?
        # self.cell(w=120, h=self.lh, txt='Features', border='B', align='C')
        self.set_font('Helvetica', '', 12)
        # n_rows = max([len(v) for v in features.values()])
        for k, v in features.items():
            self.ln(self.lh)
            # h = max(- (- sum([len(i) for i in v]) // 60), 1)
            self.cell(w=50, h=self.lh, txt=k, border='R', align='L')
            self.cell(w=50, h=self.lh, txt='{}'.format(len(v)), align='C')
            # self.multi_cell(w=140, h=self.lh, txt=', '.join([i for i in v]), align='L')
        self.check_new_page()

        # Feature Selection
        self.add_h3('Feature Selection')
        self.set_font('Helvetica', '', 12)
        y = self.get_y()
        self.multi_cell(w=80, h=self.lh,
                        txt="Using a Random Forest model, the non-linear Feature Importance is analysed. The feature "
                            "importance is measured in Mean Decrease in Gini Impurity. The feature importance is used "
                            "to create two feature sets, one that contains 85% of all feature importance, and one that "
                            "contains all features that contribute more than 1% to the total feature importance.\nThe "
                            "top 15 feature with their mean decrease in Gini impurity are visualized on the right.")
        path = self.p.mainDir + 'EDA/Features/v{}/RF.png'.format(self.p.version)
        self.image(name=path, x=110 + self.pm, y=y - 10, w=80, h=120)

    def data(self):
        if not self.check_new_page():
            self.ln(30)
        self.add_h2('Data Processing')
        self.add_text("The following data manipulations were made to clean the data:\n"
                      "  1.  Removed {} duplicate columns and {} duplicate rows\n"
                      "  2.  Removed {} outliers with {}\n"
                      "  3.  Imputed {} missing values with {}\n"
                      "  4.  Removed {} columns with constant values\n"
                      .format(self.p.dataProcesser.removedDuplicateColumns, self.p.dataProcesser.removedDuplicateRows,
                              self.p.dataProcesser.removedOutliers, self.p.dataProcesser.outlier_removal,
                              self.p.dataProcesser.imputedMissingValues, self.p.dataProcesser.missing_values,
                              self.p.dataProcesser.removedConstantColumns))

    def score_board(self):
        if not self.check_new_page():
            self.ln(self.lh)
        scores = self.p.sort_results(self.p.results)
        self.ln(self.lh)
        self.add_h2('Model Score Board')
        self.add_text("Not only the {} has been optimized by the AutoML pipeline. In total, {} models where trained. "
                      "The following table shows the performance of the top 10 performing models:"
                      .format(self.mName, len(self.p.results)))
        self.ln(self.lh)
        # Table
        self.set_font('Helvetica', 'B', 12)
        self.cell(w=70, h=self.lh, txt='Model', border='RB', align='C')
        self.cell(w=70, h=self.lh, txt=self.p.objective, border='B', align='C')
        # no space for parameters, move to appendix if wanted
        # self.cell(w=100, h=self.lh, txt='Parameters', border='B', align='C')
        self.set_font('Helvetica', '', 12)
        for i in range(min(10, len(scores))):
            self.ln(self.lh)
            self.cell(w=70, h=self.lh, txt='{}'.format(scores.iloc[i]['model']), border='R', align="C")
            self.cell(w=70, h=self.lh, txt='{:.4f} \u00B1 {:.4f} %'.format(scores.iloc[i]['mean_objective'],
                                                                           scores.iloc[i]['std_objective']), align='C')



import logging
import numpy as np
import pandas as pd
from scipy import stats
import matplotlib.pyplot as plt
from Amplo.Utils import histSearch


class DriftDetector:
    # todo add second order pdf fit
    # todo add subsequence drift detection

    def __init__(self,
                 num_cols: list = None,
                 cat_cols: list = None,
                 date_cols: list = None,
                 n_bins: int = 500,
                 sigma: int = 3,
                 with_pdf: bool = False
                 ):
        """
        Detects data drift in streamed input data.
        Supports numerical, categorical and datetime variables.
        Due to streamed, we don't check distributions, just bins.
        Categorical simply checks whether it's not a new column
        Datetime simply checks whether the date is recent
        """
        # Copy kwargs
        self.num_cols = [] if num_cols is None else num_cols
        self.cat_cols = [] if cat_cols is None else cat_cols
        self.date_cols = [] if date_cols is None else date_cols
        self.n_bins = n_bins
        self.with_pdf = with_pdf
        self.sigma = sigma

        # Initialize
        self.bins = {}
        self.distributions = {}

    def fit(self, data: pd.DataFrame) -> object:
        """
        Fits the class object
        """
        # Numerical
        self._fit_bins(data)
        self._fit_distributions(data)

        # Categorical

        return self

    def check(self, data: pd.DataFrame):
        """
        Checks a new dataframe for distribution drift.
        """
        violations = []

        # Numerical
        violations.extend(self._check_bins(data))

        return violations

    def get_weights(self) -> dict:
        """
        Gets the weights of the fitted object.
        Useful to save :)
        """
        return {
            'bins': self.bins,
            'distributions': self.distributions
        }

    def load_weights(self, weights: dict):
        """
        Sets the weights of the object to recreate a previously fitted object.

        Parameters
        ----------
        weights:
            bins (dict): Bins dictionary with bins and quantities for all numeric keys
            distributions (dict): Dictionary with fitted distributions for all numeric keys.
        """
        self.bins = weights['bins']
        self.distributions = weights['distributions']
        return self

    def _fit_bins(self, data: pd.DataFrame):
        """
        Fits a histogram on each numerical column.
        """
        for key in self.num_cols:
            y, x = np.histogram(data[key], bins=self.n_bins)
            self.bins[key] = (x.tolist(), y.tolist())

    def _check_bins(self, data: pd.DataFrame):
        """
        Checks if the current data falls into bins
        """
        violations = []

        for key in self.num_cols:
            # Get bins
            x, y = self.bins[key]

            # Check bins
            if isinstance(data, pd.DataFrame):
                for v in data[key].values:
                    ind = histSearch(x, v)
                    if ind == -1 or y[ind] <= 0:
                        violations.append(key)
                        break
            elif isinstance(data, pd.Series):
                ind = histSearch(x, data[key])
                if ind == -1 or (y[ind] <= 0 and y[min(0, ind - 1)] <= 0 and y[max(self.n_bins, ind + 1)] <= 0):
                    violations.append(key)

        if len(violations) > 0:
            logging.warning(f"[AutoML] Drift detected!  {len(violations)} features outside training bins: {violations}")

        return violations

    def _fit_distributions(self, data: pd.DataFrame):
        """
        Fits a distribution on each numerical column.
        """
        if self.with_pdf:
            distributions = ["gamma", "beta", "dweibull", "dgamma"]
            distances = []
            fitted = []

            # Iterate through numerical columns
            for key in self.num_cols:
                y, x = np.histogram(data[key], normed=True)
                x = (x + np.roll(x, -1))[:-1] / 2.0     # Get bin means

                # Iterate through distributions
                for distribution in distributions:
                    # Fit & Get PDF
                    dist = getattr(stats, distribution)

                    # Multiple order fit
                    params = dist.fit(data[key])
                    fitted_pdf = dist.pdf(x, loc=params[-2], scale=params[-1], *params[:-2])

                    # Analyse
                    distances.append(sum((y - fitted_pdf) ** 2))
                    fitted.append({
                        'distribution': distribution,
                        'params': params,
                    })
                plt.legend(['Original'] + distributions)
                plt.show()

                # Select lowest
                self.distributions[key] = fitted[np.argmin(distances)]

    def _check_distributions(self, data: pd.DataFrame) -> list:
        """
        Checks whether the new data falls within the fitted distributions
        """
        if self.with_pdf:
            # Init
            violations = []

            # Check all numerical columns
            for key in self.num_cols:
                dist = getattr(stats, self.distributions[key]['distribution'])
                params = self.distributions[key]['params']
                probabilities = dist.pdf(data[key].values, loc=params[-2], scale=params[-1], *params[:-2])

                if any(p < self.sigma for p in probabilities):
                    violations.append(key)
                    continue

            if len(violations) > 0:
                logging.warning(f"[AutoML] Drift detected!  {len(violations)} features outside training bins: "
                                f"{violations}")

            return violations



import unittest
import pandas as pd
from sklearn.datasets import make_regression
from sklearn.base import clone
from sklearn.metrics import SCORERS
from Amplo.Regressors import StackingRegressor


class TestStackingRegression(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cls.model = StackingRegressor()
        x, y = make_regression()
        cls.x, cls.y = pd.DataFrame(x), pd.Series(y)

    def test_set_params(self):
        self.model.set_params(**{'CatBoostRegressor': {'depth': 10}})

    def test_get_params(self):
        self.model.get_params()

    def test_pd_fit(self):
        model = clone(self.model)
        model.fit(self.x, self.y)

    def test_np_fit(self):
        model = clone(self.model)
        model.fit(self.x.to_numpy(), self.y.to_numpy())

    def test_trained_attr(self):
        model = clone(self.model)
        assert hasattr(model, 'trained')
        assert model.trained is False
        model.fit(self.x, self.y)
        assert model.trained is True

    def test_predict(self):
        model = clone(self.model)
        model.fit(self.x, self.y)
        prediction = model.predict(self.x)

        assert len(prediction.shape) == 1

    def test_cloneable(self):
        model = clone(self.model)
        params = {'CatBoostRegressor': {'depth': 10}}
        cloned = clone(clone(model).set_params(**params))
        cloned.fit(self.x, self.y)
        prediction = cloned.predict(self.x)

        assert len(prediction.shape) == 1

    def test_scorer(self):
        model = clone(self.model)
        model.fit(self.x, self.y)
        SCORERS['neg_mean_absolute_error'](model, self.x, self.y)
        SCORERS['neg_mean_squared_error'](model, self.x, self.y)
        SCORERS['r2'](model, self.x, self.y)




import pytest
import unittest
import numpy as np
import pandas as pd
from scipy import stats
from Amplo.AutoML import DriftDetector
import matplotlib.pyplot as plt


def draw(n):
    return pd.DataFrame({
        'norm': np.random.normal(0.23, 0.53, n),
        'uniform': np.random.uniform(0, 100, n),
        'exponential': np.random.exponential(0.4, n),
        'gamma': np.random.gamma(0.3, 0.9, n),
        'beta': np.random.beta(0.2, 0.4, n),
    })


class TestDriftDetector(unittest.TestCase):

    def test_distribution_fits(self):
        # Setup
        dists = ["norm", "uniform", "expon", "gamma", "beta"]
        ref = draw(500)
        test = ref.iloc[np.random.permutation(len(ref))[:10]]
        drift = DriftDetector(num_cols=ref.keys())
        drift.fit(ref)

        # Checks
        assert len(drift.check(test)) == 0, "Test data found inconsistent"
        assert len(drift.check(ref.max() + 1)) == len(dists), "Maxima not detected"
        assert len(drift.check(ref.min() - 1)) == len(dists), "Minima not detected"



import re
import json
import pandas as pd
from typing import Union


def influx_query_to_df(result):
    df = []
    for table in result:
        parsed_records = []
        for record in table.records:
            parsed_records.append((record.get_time(), record.get_value()))
        df.append(pd.DataFrame(parsed_records, columns=['ts', record.get_field()]))
    return pd.concat(df).set_index('ts').groupby(level=0).sum()


def getModel(model_str, **args):
    from Amplo.AutoML.Modeller import Modeller
    try:
        return [mod for mod in Modeller(**args).return_models() if type(mod).__name__ == model_str][0]
    except IndexError:
        raise IndexError('Model not found.')


def histSearch(array, value):
    low = 0
    high = len(array) - 1
    while low < high:
        middle = low + (high - low) // 2
        if value > array[middle + 1]:
            low = middle + 1
        elif value < array[middle]:
            high = middle
        else:
            return middle
    return -1


def boolean_input(question: str) -> bool:
    x = input(question + ' [y / n]')
    if x.lower() == 'n' or x.lower() == 'no':
        return False
    elif x.lower() == 'y' or x.lower() == 'yes':
        return True
    else:
        print('Sorry, I did not understand. Please answer with "n" or "y"')
        return boolean_input(question)


def clean_keys(data: pd.DataFrame) -> pd.DataFrame:
    # Clean Keys
    new_keys = {}
    for key in data.keys():
        if isinstance(key, int):
            new_keys[key] = 'feature_{}'.format(key)
        else:
            new_keys[key] = re.sub('[^a-z0-9\n]', '_', str(key).lower()).replace('__', '_')
    data = data.rename(columns=new_keys)
    return data


def parse_json(json_string: Union[str, dict]) -> Union[str, dict]:
    if isinstance(json_string, dict):
        return json_string
    else:
        try:
            return json.loads(json_string
                              .replace("'", '"')
                              .replace("True", "true")
                              .replace("False", "false")
                              .replace("nan", "NaN")
                              .replace("None", "null"))
        except json.decoder.JSONDecodeError:
            print('[AutoML] Cannot validate, impassable JSON.')
            print(json_string)
            return json_string


def check_dataframe_quality(data: pd.DataFrame) -> bool:
    assert not data.isna().any().any()
    assert not data.isnull().any().any()
    assert not (data.dtypes == object).any().any()
    assert not (data.dtypes == str).any().any()
    assert data.max().max() < 1e38 and data.min().min() > -1e38
    return True



import unittest
import numpy as np
import pandas as pd
from sklearn.datasets import make_classification
from sklearn.datasets import make_regression
from Amplo.AutoML import FeatureProcesser


class TestDataProcessing(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        pass

    def test_regression(self):
        x, y = make_regression()
        x, y = pd.DataFrame(x), pd.Series(y)
        fp = FeatureProcesser(max_lags=2, mode='regression')
        xt, sets = fp.fit_transform(x, y)

    def test_classification(self):
        x, y = make_classification()
        x, y = pd.DataFrame(x), pd.Series(y)
        fp = FeatureProcesser(max_lags=2, mode='classification')
        xt, sets = fp.fit_transform(x, y)

    def test_co_linearity(self):
        y = pd.Series(np.linspace(2, 100, 100))
        x = pd.DataFrame({'a': np.linspace(-4, 4, 100), 'b': np.linspace(-4, 4, 100)})
        fp = FeatureProcesser(mode='regression')
        xt, sets = fp.fit_transform(x, y)
        assert len(fp.coLinearFeatures) != 0, "Colinear feature not removed"

    def test_multiply_features(self):
        y = pd.Series(np.linspace(2, 100, 100))
        b = pd.Series(np.linspace(-4, 4, 100) ** 2)
        x = pd.DataFrame({'a': y / b, 'b': b})
        fp = FeatureProcesser(mode='regression')
        xt, sets = fp.fit_transform(x, y)
        assert len(fp.crossFeatures) != 0, "Multiplicative feature not spotted"

    def test_division(self):
        y = pd.Series(np.linspace(2, 100, 100))
        b = pd.Series(np.linspace(-4, 4, 100) ** 2)
        x = pd.DataFrame({'a': y * b, 'b': b})
        fp = FeatureProcesser(mode='regression')
        xt, sets = fp.fit_transform(x, y)
        assert len(fp.crossFeatures) != 0, "Division feature not spotted"

    def test_trigonometry(self):
        y = pd.Series(np.sin(np.linspace(0, 100, 100)))
        x = pd.DataFrame({'a': np.linspace(0, 100, 100)})
        fp = FeatureProcesser(mode='regression')
        xt, sets = fp.fit_transform(x, y)
        assert len(fp.trigonometricFeatures) != 0, "Trigonometric feature not spotted"

    # def test_inverse(self): --> Invariant for decision tree
    #     y = pd.Series(np.random.randint(1, 100, 100))
    #     x = pd.DataFrame({'a': 1 / y})
    #     print(y, x)
    #     fp = FeatureProcessing(mode='regression')
    #     xt, sets = fp.fit_transform(x, y)
    #     assert len(fp.inverseFeatures) != 0, "Inverse feature not spotted"

    def test_lagged(self):
        y = pd.Series(np.random.randint(0, 100, 100))
        x = pd.DataFrame({'a': np.roll(y, -5)})
        fp = FeatureProcesser(mode='regression')
        xt, sets = fp.fit_transform(x, y)
        assert len(fp.laggedFeatures) != 0, "Lagged feature not spotted"

    def test_diff(self):
        y = pd.Series(np.random.randint(1, 100, 100))
        x = pd.DataFrame({'a': np.cumsum(y)})
        fp = FeatureProcesser(mode='regression')
        xt, sets = fp.fit_transform(x, y)
        assert len(fp.diffFeatures) != 0, "Difference feature not spotted"

    def test_select(self):
        y = pd.Series(np.linspace(0, 100, 100))
        x = pd.DataFrame({'a': y, 'b': np.random.randint(0, 100, 100)})
        fp = FeatureProcesser(mode='regression')
        xt, sets = fp.fit_transform(x, y)
        assert all([len(i) == 1 for i in sets.values()]), "Random Feature Selected"

    def test_settings(self):
        y = pd.Series(np.random.randint(1, 100, 100))
        b = pd.Series(np.linspace(-4, 4, 100))
        x = pd.DataFrame({'a': np.cumsum(y), 'b': np.roll(y, -5), 'c': y / b, 'd': y * b})
        fp = FeatureProcesser(mode='regression')
        xt, sets = fp.fit_transform(x, y)
        settings = fp.get_settings()
        fpn = FeatureProcesser(mode='regression')
        fpn.load_settings(settings)
        for k, v in sets.items():
            xtn = fpn.transform(x, k)
            assert len(v) == len(xtn.keys()), "Incorrect number of keys"
            assert all(xt[v].keys() == xtn.keys()), 'Keys are not correct'
            assert np.allclose(xt[v], xtn), 'Transformed data not consistent for {} set'.format(k)



import unittest
import pandas as pd
from sklearn.datasets import make_regression
from Amplo.Regressors import LGBMRegressor
from .test_regressor import TestRegression


class TestLGBMRegressor(unittest.TestCase, TestRegression):

    @classmethod
    def setUpClass(cls):
        cls.model = LGBMRegressor()
        x, y = make_regression(n_informative=15)
        cls.x, cls.y = pd.DataFrame(x), pd.Series(y)



import os
import numpy as np
import pandas as pd
from annoy import AnnoyIndex
from Amplo.AutoML import DataProcesser


class IntervalAnalyser:

    def __init__(self,
                 folder: str = None,
                 norm: str = 'euclidean',
                 n_neighbors: int = None,
                 n_trees: int = 10,
                 ):
        """
        Interval Analyser for Log file classification. When log files have to be classifier, and there is not enough
        data for time series methods (such as LSTMs, ROCKET or Weasel, Boss, etc), one needs to fall back to classical
        machine learning models which work better with lower samples. But this raises the problem of which samples to
        classify. You shouldn't just simply classify on every sample and accumulate, that may greatly disrupt
        classification performance. Therefore, we introduce this interval analyser. By using K-Nearest Neighbors,
        one can estimate the strength of correlation for every sample inside a log. Using this allows for better
        interval selection for classical machine learning models.

        To use this interval analyser, make sure:
        - That your logs are of equal length
        - That your logs have identical keys
        - That your logs are located in a folder of their class, with one parent folder with all classes, e.g.:
        +-- Parent Folder
        |   +-- Class_1
        |       +-- Log_1.*
        |       +-- Log_2.*
        |   +-- Class_2
        |       +-- Log_3.*

        Parameters
        ----------
        folder [str]:       Parent folder of classes
        index_col [str]:    For reading the log files
        norm [str]:         Optimization metric for K-Nearest Neighbors
        n_neighbors [int]:    Quantity of neighbors, default to 3 * log length
        n_trees [int]:      Quantity of trees
        """
        # Parameters
        self.folder = folder + '/' if len(folder) == 0 or folder[-1] != '/' else folder
        self.norm = norm
        self.n_trees = n_trees
        self.n_samples, self.n_keys = self._check_length()
        self.n_neighbors = 3 * self.n_samples if n_neighbors is None else n_neighbors

        # Initializers
        self.n_files = 0
        self.n_folders = 0
        self.means = None
        self.stds = None

        # Test
        assert norm in ['euclidean', 'manhattan', 'angular', 'hamming', 'dot']

    def analyse(self) -> np.ndarray:
        """
        Function that runs the K-Nearest Neighbors and returns a NumPy array with the sensitivities.

        Returns
        -------
        np.array: Estimation of strength of correlation
        """
        # Set up
        engine = AnnoyIndex(self.n_keys, self.norm)

        # Get data
        df, labels = self._parse_data()

        # Index
        for i, row in df.iterrows():
            engine.add_item(i[0] * self.n_samples + i[1], row.values)

        # Build
        engine.build(self.n_trees)

        # Return distribution
        return self._make_distribution(engine, df, labels)

    def save(self, path):
        pass

    def load(self, path):
        pass

    def _check_length(self) -> [int, int]:
        """
        Detects the extension from the log files
        """
        # Read first log
        folder = os.listdir(self.folder)[0]
        file = os.listdir(self.folder + folder)[0]
        df = self._read(f"{self.folder}{folder}/{file}")

        # Set lengths
        length, keys = len(df), len(df.keys())

        return length, keys

    def _read(self, path: str) -> pd.DataFrame:
        """
        Wrapper for various read functions
        """
        f_ext = path[path.rfind('.'):]
        if f_ext == '.csv':
            return pd.read_csv(path)
        elif f_ext == '.json':
            return pd.read_json(path)
        elif f_ext == '.xml':
            return pd.read_xml(path)
        elif f_ext == '.feather':
            return pd.read_feather(path)
        elif f_ext == '.parquet':
            return pd.read_parquet(path)
        elif f_ext == '.stata':
            return pd.read_stata(path)
        elif f_ext == '.pickle':
            return pd.read_pickle(path)
        else:
            raise NotImplementedError('File format not supported.')

    def _parse_data(self) -> [pd.DataFrame, pd.Series]:
        """
        Reads all files and sets a multi index
        Returns
        -------
        pd.DataFrame: all log files
        """
        # Result init
        dfs = []

        # Loop through files
        for folder in os.listdir(self.folder):
            for file in os.listdir(self.folder + folder):

                # Read df
                df = self._read(f'{self.folder}{folder}/{file}')

                # Set label
                df['class'] = folder

                # Set second index
                df = df.set_index(pd.MultiIndex.from_product([[self.n_files], df.index.values], names=['log', 'index']))

                # Add to list
                dfs.append(df)

                # Increment
                self.n_files += 1
            self.n_folders += 1

        # Concatenate dataframes
        dfs = pd.concat(dfs)

        # Remove classes
        classes = dfs['class']
        dfs = dfs.drop('class', axis=1)
        dp = DataProcesser(missing_values='zero')
        dfs = dp.fit_transform(dfs)

        # Return
        return dfs, classes

    def _make_distribution(self, engine, df: pd.DataFrame, classes: pd.Series) -> np.ndarray:
        """
        Given a build K-Nearest Neighbors, returns the label distribution
        """
        # Result setup
        distribution = [[] for _ in range(self.n_files)]

        # Iterate through samples
        for (log, sample), data in df.iterrows():
            # Get neighbor indices
            neighbor_inds = engine.get_nns_by_vector(data.values, self.n_neighbors)

            # Set percentage
            distribution[log].append((classes.iloc[neighbor_inds] == classes[(log, sample)]).sum() / self.n_neighbors)

        # Mask
        lens = np.array([len(item) for item in distribution])
        mask = lens[:, None] > np.arange(lens.max())
        out = np.full(mask.shape, 0)
        out[mask] = np.concatenate(distribution)
        distribution = np.array(out)

        # Average over logs
        return np.mean(distribution * 100, axis=0)



import time
import copy
import random
import warnings
import itertools
import numpy as np
import pandas as pd
from tqdm import tqdm
import multiprocessing as mp
from datetime import datetime
from sklearn.metrics import SCORERS


# noinspection PyUnresolvedReferences
class BaseGridSearch:

    def __init__(self, model, params=None, cv=None, scoring='accuracy',
                 timeout=3600, candidates=250, verbose=0):
        """
        Basic exhaustive grid search.
        @param model: Model object to optimize
        @param cv: Cross-Validator for scoring
        @param scoring: make_scorer for scoring
        """
        # Args
        self.model = model
        self.params = params
        self.cv = cv
        self.scoring = SCORERS[scoring] if isinstance(scoring, str) else scoring
        self.timeout = timeout
        self.nTrials = candidates
        self.verbose = verbose
        self.mode = 'regression' if 'Regressor' in type(model).__name__ else 'classification'
        self.binary = True
        self.samples = None

        # Checks
        assert model is not None, 'Model not provided'
        assert self.scoring in SCORERS.values(), 'Chose a scorer from sklearn.metrics.SCORERS'
        
        # Initiate
        self.parsedParams = []
        self.result = []

    def _get_hyper_params(self):
        # Parameters for both Regression / Classification
        if type(self.model).__name__ == 'LinearRegression':
            return {}
        elif type(self.model).__name__ == 'Lasso' or \
                'Ridge' in type(self.model).__name__:
            return {
                'alpha': np.logspace(-5, 0, 25).tolist(),
            }
        elif 'SV' in type(self.model).__name__:
            return {
                'gamma': ['scale', 'auto', 0.001, 0.01, 0.1, 0.5, 1],
                'C': np.logspace(-5, 0, 25).tolist(),
            }
        elif 'KNeighbors' in type(self.model).__name__:
            return {
                'n_neighbors': np.linspace(5, min(50, int(self.samples / 10)), 5).astype('int').tolist(),
                'weights': ['uniform', 'distance'],
                'leaf_size': np.linspace(1, min(150, int(self.samples / 10), 5).astype('int').tolist()),
                'n_jobs': [mp.cpu_count() - 1],
            }
        elif 'MLP' in type(self.model).__name__:
            return {
                'hidden_layer_sizes': [(100,), (100, 100), (100, 50), (200, 200), (200, 100), (200, 50),
                                       (50, 50, 50, 50)],
                'learning_rate': ['adaptive', 'invscaling'],
                'alpha': [1e-4, 1e-3, 1e-5],
                'shuffle': [False],
            }

        # Regressor specific hyper parameters
        elif self.mode == 'regression':
            if type(self.model).__name__ == 'SGDRegressor':
                return {
                    'loss': ['squared_loss', 'huber', 'epsilon_insensitive', 'squared_epsilon_insensitive'],
                    'penalty': ['l2', 'l1', 'elasticnet'],
                    'alpha': np.logspace(-5, 0, 5).tolist(),
                }
            elif type(self.model).__name__ == 'DecisionTreeRegressor':
                return {
                    'criterion': ['mse', 'friedman_mse', 'mae', 'poisson'],
                    'max_depth': np.linspace(3, min(15, int(np.log2(self.samples))), 4).astype('int').tolist(),
                }
            elif type(self.model).__name__ == 'AdaBoostRegressor':
                return {
                    'n_estimators': [25, 50, 75, 100, 150, 200, 250],
                    'loss': ['linear', 'square', 'exponential'],
                    'learning_rate': np.logspace(-5, 0, 10).tolist()
                }
            elif type(self.model).__name__ == 'BaggingRegressor':
                return {
                    # 'n_estimators': [5, 10, 15, 25, 50],
                    'max_features': [1, 0.9, 0.8, 0.5],
                    'max_samples': [1, 0.9, 0.8, 0.5],
                    'bootstrap': [False, True],
                    'bootstrap_features': [True, False],
                    'n_jobs': [mp.cpu_count() - 1],
                }
            elif type(self.model).__name__ == 'GradientBoostingRegressor':
                return {
                    'loss': ['ls', 'lad', 'huber'],
                    'learning_rate': np.logspace(-5, 0, 10).tolist(),
                    'max_depth': np.linspace(3, min(15, int(np.log2(self.samples))), 4).astype('int').tolist(),
                }
            elif type(self.model).__name__ == 'HistGradientBoostingRegressor':
                return {
                    'max_iter': [100, 250],
                    'max_bins': [100, 150, 200, 255],
                    'loss': ['least_squares', 'least_absolute_deviation'],
                    'l2_regularization': np.logspace(-5, 0, 5).tolist(),
                    'learning_rate': np.logspace(-5, 0, 5).tolist(),
                    'max_leaf_nodes': [30, 50, 100, 150, 200, 250],
                    'early_stopping': [True],
                }
            elif type(self.model).__name__ == 'RandomForestRegressor':
                return {
                    'criterion': ['mse', 'mae'],
                    'max_depth': np.linspace(3, min(15, int(np.log2(self.samples))), 4).astype('int').tolist(),
                    'max_features': ['auto', 'sqrt'],
                    'min_samples_split': [2, 25, 50],
                    'min_samples_leaf': np.logspace(0, min(3, int(np.log10(100)) - 1), 5).astype('int').tolist(),
                    'bootstrap': [True, False],
                }
            elif type(self.model).__name__ == 'CatBoostRegressor':
                return {
                    'verbose': [0],
                    'loss_function': ['MAE', 'RMSE'],
                    'allow_writing_files': [False],
                    'early_stopping_rounds': [100],
                    'learning_rate': np.logspace(-5, 0, 5).tolist(),
                    'l2_leaf_reg': np.logspace(-5, 0, 5).tolist(),
                    'depth': [3, 5, 7, 10],
                    'min_data_in_leaf': np.logspace(0, min(3, int(np.log10(100)) - 1), 5).astype('int').tolist(),
                    'grow_policy': ['SymmetricTree', 'Depthwise', 'Lossguide'],
                }
            elif type(self.model).__name__ == 'XGBRegressor':
                return {
                    'max_depth': np.linspace(3, min(15, int(np.log2(self.samples))), 4).astype('int').tolist(),
                    'booster': ['gbtree', 'gblinear', 'dart'],
                    'learning_rate': np.logspace(-5, 0, 10).tolist(),
                    'verbosity': [0],
                    'n_jobs': [mp.cpu_count() - 1],
                }
            elif type(self.model).__name__ == 'LGBMRegressor':
                return {
                    'verbosity': [-1],
                    'num_leaves': [10, 50, 100, 150],
                    'min_child_samples': [1, 50, 500, 1000],
                    'min_child_weight': np.logspace(-5, 0, 5).tolist(),
                    'subsample': [1, 0.9, 0.8, 0.5],
                    'colsample_bytree': [1, 0.9, 0.8, 0.5],
                    'reg_alpha': np.logspace(-5, 0, 5).tolist(),
                    'reg_lambda': np.logspace(-5, 0, 5).tolist(),
                    'n_jobs': [mp.cpu_count() - 1],
                }

        # Classification specific hyper parameters
        elif self.mode == 'classification':
            if type(self.model).__name__ == 'SGDClassifier':
                return {
                    'loss': ['hinge', 'log', 'modified_huber', 'squared_hinge'],
                    'penalty': ['l2', 'l1', 'elasticnet'],
                    'alpha': np.logspace(-5, 0, 5).tolist(),
                    'max_iter': [250, 1000, 1500],
                }
            elif type(self.model).__name__ == 'DecisionTreeClassifier':
                return {
                    'criterion': ['gini', 'entropy'],
                    'max_depth': np.linspace(3, min(50, int(np.log2(self.samples))), 4).astype('int').tolist(),
                }
            elif type(self.model).__name__ == 'AdaBoostClassifier':
                return {
                    'n_estimators': [25, 50, 100, 150, 250],
                    'learning_rate': np.logspace(-5, 0, 10).tolist()
                }
            elif type(self.model).__name__ == 'BaggingClassifier':
                return {
                    # 'n_estimators': [5, 10, 15, 25, 50],
                    'max_features': [1, 0.9, 0.8, 0.5],
                    'max_samples': [1, 0.9, 0.8, 0.5],
                    'bootstrap': [False, True],
                    'bootstrap_features': [True, False],
                    'n_jobs': [mp.cpu_count() - 1],
                }
            elif type(self.model).__name__ == 'GradientBoostingClassifier':
                return {
                    'loss': ['deviance', 'exponential'],
                    'learning_rate': np.logspace(-5, 0, 10).tolist(),
                    'max_depth': np.linspace(3, min(15, int(np.log2(self.samples))), 4).astype('int').tolist(),
                }
            elif type(self.model).__name__ == 'HistGradientBoostingClassifier':
                return {
                    'max_iter': [100, 250, 500, 1000],
                    'max_bins': [100, 175, 255],
                    'l2_regularization': np.logspace(-5, 0, 5).tolist(),
                    'learning_rate': np.logspace(-5, 0, 5).tolist(),
                    'max_leaf_nodes': [30, 100, 200],
                    'early_stopping': [True]
                }
            elif type(self.model).__name__ == 'RandomForestClassifier':
                return {
                    'criterion': ['gini', 'entropy'],
                    'max_depth': np.linspace(3, min(15, int(np.log2(self.samples))), 4).astype('int').tolist(),
                    'max_features': ['auto', 'sqrt'],
                    'min_samples_split': [2, 50, 500],
                    'min_samples_leaf': np.logspace(0, min(3, int(np.log10(100)) - 1), 5).astype('int').tolist(),
                    'bootstrap': [True, False],
                }
            elif type(self.model).__name__ == 'CatBoostClassifier':
                return {
                    'verbose': [0],
                    'loss_function': ['Logloss'] if self.binary else ['MultiClass'],
                    'allow_writing_files': [False],
                    'early_stopping_rounds': [100],
                    'learning_rate': np.logspace(-5, 0, 5).tolist(),
                    'l2_leaf_reg': np.logspace(-5, 0, 5).tolist(),
                    'depth': [3, 5, 7, 10],
                    'min_data_in_leaf': np.logspace(0, min(3, int(np.log10(100)) - 1), 5).astype('int').tolist(),
                    'grow_policy': ['SymmetricTree', 'Depthwise', 'Lossguide'],
                }
            elif type(self.model).__name__ == 'XGBClassifier':
                return {
                    "objective": ["binary:logistic"] if self.binary else ['multi:softprob'],
                    'max_depth': np.linspace(3, min(15, int(np.log2(self.samples))), 4).astype('int').tolist(),
                    'booster': ['gbtree', 'gblinear', 'dart'],
                    'learning_rate': np.logspace(-5, 0, 10).tolist(),
                    'verbosity': [0],
                    'n_jobs': [mp.cpu_count() - 1],
                    'scale_pos_weight': [0, 10, 50, 100],
                }
            elif type(self.model).__name__ == 'LGBMClassifier':
                return {
                    "objective": ["binary"] if self.binary else ['multiclass'],
                    'verbosity': [-1],
                    'num_leaves': [10, 50, 100, 150],
                    'min_child_samples': [1, 50, 500, 1000],
                    'min_child_weight': np.logspace(-5, 0, 5).tolist(),
                    'subsample': np.logspace(-5, 0, 5).tolist(),
                    'colsample_bytree': np.logspace(-5, 0, 5).tolist(),
                    'reg_alpha': np.logspace(-5, 0, 5).tolist(),
                    'reg_lambda': np.logspace(-5, 0, 5).tolist(),
                    'n_jobs': [mp.cpu_count() - 1],
                }

        # Raise error if nothing is returned
        warnings.warn('Hyper parameter tuning not implemented for {}'.format(type(self.model).__name__))
        return {}

    def _parse_params(self):
        if len(self.params.items()) > 0:
            k, v = zip(*self.params.items())
            self.parsedParams = [dict(zip(k, v)) for v in itertools.product(*self.params.values())]
            random.shuffle(self.parsedParams)
            print('[GridSearch] Optimizing {}, {}-folds with {} parameter combinations, {} runs.' .format(
                type(self.model).__name__, self.cv.n_splits, len(self.parsedParams),
                len(self.parsedParams) * self.cv.n_splits))
        else:
            self.parsedParams = [{}]

    def fit(self, x, y):
        start_time = time.time()
        # todo check data
        # Parse Params
        if self.mode == 'classification':
            if len(np.unique(y)) == 2:
                self.binary = True
            else:
                self.binary = False
        # Convert to Numpy
        if isinstance(x, pd.DataFrame) or isinstance(x, pd.Series):
            x = np.array(x)
        if isinstance(y, pd.DataFrame) or isinstance(y, pd.Series):
            y = np.array(y).reshape((-1))
        self.samples = len(y)

        # Get params
        if self.params is None:
            self.params = self._get_hyper_params()
        self._parse_params()

        # Loop through parameters
        for i, param in tqdm(enumerate(self.parsedParams)):
            # Timeout or trials limit
            if time.time() - start_time > self.timeout or i > self.nTrials:
                return pd.DataFrame(self.result)

            # Scoring and timing vector
            scoring = []
            timing = []

            # Cross Validation
            for train_ind, val_ind in self.cv.split(x, y):
                # Start Timer
                t = time.time()

                # Split data
                x_train, x_val = x[train_ind], x[val_ind]
                y_train, y_val = y[train_ind], y[val_ind]

                # Model training
                model = copy.deepcopy(self.model)
                model.set_params(**param)
                model.fit(x_train, y_train)

                # Results
                scoring.append(self.scoring(model, x_val, y_val))
                timing.append(time.time() - t)

            # Output Printing
            if self.verbose > 0:
                print('[GridSearch][{}] Score: {:.4f} \u00B1 {:.4f} (in {:.1f} seconds) (Best score so-far: {:.4f}'
                      ' \u00B1 {:.4f}) ({} / {})'.format(datetime.now().strftime('%H:%M'), np.mean(scoring),
                                                         np.std(scoring), np.mean(timing), self.best[0], self.best[1],
                                                         i + 1, len(self.parsedParams)))

            self.result.append({
                'date': datetime.today().strftime('%d %b %y'),
                'model': type(model).__name__,
                'mean_objective': np.mean(scoring),
                'std_objective': np.std(scoring),
                'worst_case': np.mean(scoring) - np.std(scoring),
                'params': param,
                'mean_time': np.mean(timing),
                'std_time': np.std(timing),
            })
        return pd.DataFrame(self.result)
